warning('off','MATLAB:table:ModifiedAndSavedVarnames')

% Read swaps from spreadsheet
swapFile = 'data/IRS_Nordea.xlsx';
cdsFile = 'data/CDS_Nordea.xlsx';

swaps = readtable(swapFile);
swaps.LegType = [swaps.LegType ~swaps.LegType];
swaps.LegRate = [swaps.LegRateReceiving swaps.LegRatePaying];
swaps.LegReset = ones(size(swaps,1),1);
swaps.Maturity = datenum(swaps.Maturity);

numSwaps = size(swaps,1);

% Create RateSpec from the Interest-Rate Curve

Settle = datenum('1-March-2019');

Tenor = [3 6 12 5*12 7*12 10*12]';
ZeroRates = [0.0137 0.0146 0.01806 0.01902 0.01983 0.02092]';

ZeroDates = datemnth(Settle,Tenor);
Compounding = 4;
Basis = 0;
RateSpec = intenvset('StartDates', Settle,'EndDates', ZeroDates, 'Rates', ZeroRates,'Compounding',Compounding,'Basis',Basis);

% Monte carlo

N = 1000;
simulationDates = datemnth(Settle,0:12);
simulationDates = [simulationDates datemnth(simulationDates(end),3:3:74)]';
numDates = numel(simulationDates);

% Compute Floating Reset Dates

floatDates = cfdates(Settle-360, swaps.Maturity, swaps.Period);
swaps.FloatingResetDates = zeros(numSwaps,numDates);
for i = numDates:-1:1
    thisDate = simulationDates(i);
    floatDates(floatDates > thisDate) = 0;
    swaps.FloatingResetDates(:,i) = max(floatDates,[],2);
end

% Setup Hull-White Single Factor Model

Alpha = 0.2;
Sigma = 0.015;

hw1 = HullWhite1F(RateSpec, Alpha, Sigma);

% Simulate scenarios

% Use reproducible random number generator (vary the seed to produce different random scenarios).
prevRNG = rng(0, 'twister');

dt = diff(yearfrac(Settle,simulationDates,1));
nPeriods = numel(dt);
scenarios = hw1.simTermStructs(nPeriods, 'nTrials', N, 'deltaTime', dt);

% Restore random number generator state
rng(prevRNG);

% Compute the discount factors through each realized interest rate scenario.
dfactors = ones(numDates, N);
for i = 2:numDates
    tenorDates = datemnth(simulationDates(i-1),Tenor);
    rateAtNextSimDate = interp1(tenorDates,squeeze(scenarios(i-1,:,:)), simulationDates(i),'linear','extrap');
    % Compute D(t1,t2)
    dfactors(i,:) = zero2disc(rateAtNextSimDate, repmat(simulationDates(i), 1, N),simulationDates(i-1), -1,3);
end
dfactors = cumprod(dfactors,1);

% Compute all mark-to-market values for this scenario
values = hcomputeMTMValues(swaps, simulationDates, scenarios,Tenor);

[exposures, expcpty] = creditexposures(values, swaps.CounterpartyID, 'NettingID',swaps.NettingID);

% Compute entire portfolio exposure
portExposures = sum(exposures, 2);

% Compute exposure profiles for entire portfolio
cpProfiles = exposureprofiles(simulationDates,exposures);
portProfiles = exposureprofiles(simulationDates,portExposures);

% Get discounted exposures for counterparty, for each scenario
discExp = zeros(size(exposures));
for i = 1:N
    discExp(:,:,i) = bsxfun(@times,dfactors(:,i),exposures(:,:,i));
end

% Discounted expected exposure
discProfiles = exposureprofiles(simulationDates, discExp, 'ProfileSpec','EE');
discEE = [discProfiles.EE];

CDS = readtable(cdsFile);
CDSDates = datenum(CDS.Date);
MarketDates = [datemnth(CDSDates(1),0:3:108)]';

CDSSpreads = table2array(CDS(:,2:end));

ZeroData = [RateSpec.EndDates RateSpec.Rates];    

% Calibrate default probabilities for each risk level
DefProb = zeros(length(MarketDates), 5);
for i = 1:size(DefProb,2)
    probData = cdsbootstrap(ZeroData, [CDSDates CDSSpreads(:,i)], Settle, 'probDates', MarketDates);
    DefProb(:,i) = probData(:,2);
end


Recovery = 0.4;
CVA = (1-Recovery) * sum(discEE(2:end,:) .* diff(DefProb));
CVA = round(CVA, 2);
% PLOT GRAPHS

% Yield curve at settle date
% subplot(4,4,1);
figure;
plot(ZeroDates, ZeroRates, 'o-');
xlabel('Date');
datetick('keeplimits');
ylabel('Zero rate');
grid on;
title('Yield Curve at Settle Date');

% Yield curve evolution for specific scenario
scenario = 10;

% subplot(4,4,2);
figure;
surf(Tenor, simulationDates, scenarios(:,:,scenario))
axis tight
datetick('y','mmmyy');
xlabel('Tenor (Months)');
ylabel('Observation Date');
zlabel('Rates');
ax = gca;
ax.View = [-49 32];
title(sprintf('Scenario %d Yield Curve Evolution\n',scenario));


% Rate Simulation for specific rate
% subplot(4,4,2);
figure;
plot(simulationDates, squeeze(scenarios(:,1,:)), '-');
xlabel('Observation Date');
datetick('keeplimits');
ylabel('Rate');
grid on;
title('Hull-White Short Rate Simulation');

% Swap prices along specific scenario
% subplot(4,4,3);
figure;
plot(simulationDates, values(:,:,scenario));
datetick('x','mmmyy','keeplimits')
ylabel('Mark-To-Market Price');
title(sprintf('Swap prices along scenario %d', scenario));

% View portfolio value over time
% subplot(4,4,4);
figure;
totalPortValues = squeeze(sum(values, 2));
plot(simulationDates, totalPortValues);
title('Total MTM Portfolio Value for All Scenarios');
datetick('x','mmmyy','keeplimits')
ylabel('Portfolio Value (NOK)')
xlabel('Simulation Dates')

% Portfolio exposure for all scenarios
% subplot(4,4,5);
figure;
totalPortExposure = squeeze(sum(exposures, 2));
plot(simulationDates, totalPortExposure);
title('Portfolio Exposure for All Scenarios');
datetick('x','mmmyy')
ylabel('Exposure (NOK)')
xlabel('Simulation Dates')

% Portfolio discounted EE
% subplot(4,4,6);
figure;
plot(simulationDates,sum(discEE,2))
datetick('x','mmmyy','keeplimits')
title('Discounted Expected Exposure for Portfolio');
ylabel('Discounted Exposure (NOK)')
xlabel('Simulation Dates')

% Plot of the cumulative probability of default for each risk level.
% subplot(4,4,8);
figure;
plot(MarketDates,DefProb)
legend('Low risk', 'Medium risk', 'High risk', 'Constant risk', 'Drastic change')
title('Default Probability Curve for Each Risk Level');
xlabel('Date');
grid on;
ylabel('Cumulative Probability')
datetick('x','mmmyy')
ylabel('Probability of Default')
xlabel('Simulation Dates')

% CVA for each risk level
% subplot(4,4,9);
figure;
bar(CVA);
ax=gca;
ax.YTickLabel=cellstr(num2str(ax.YTick'));
labels = {'Low risk', 'Medium risk', 'High risk', 'Constant risk', 'Drastic change'};
set(gca,'xticklabel',labels)
text(1:length(CVA),CVA,num2str(CVA'),'vert','bottom','horiz','center'); 
title('CVA for each risk level');
xlabel('Risk level');
ylabel('CVA NOK');
grid on;

% Portfolio Exposure Profiles

figure;
plot(simulationDates,portProfiles.PFE, ...
    simulationDates,portProfiles.MPFE * ones(numDates,1), ...
    simulationDates,portProfiles.EE, ...
    simulationDates,portProfiles.EPE * ones(numDates,1), ...
    simulationDates,portProfiles.EffEE, ...
    simulationDates,(portProfiles.EffEPE * 1.4) * ones(numDates,1));
legend({'PFE (95%)','Max PFE','Exp Exposure (EE)','Time-Avg EE (EPE)', ...
    'Max past EE (EffEE)','Time-Avg EffEE (EffEPE)'})

datetick('x','mmmyy')
title('Portfolio Exposure Profiles');
ylabel('Exposure NOK')
xlabel('Simulation Dates')

